# download data from single cell repos

import logging, optparse, io, sys, os, shutil, operator
from collections import defaultdict, OrderedDict, Counter

from .cellbrowser import runGzip, openFile, errAbort, setDebug, moveOrGzip, makeDir, iterItems
from .cellbrowser import mtxToTsvGz, writeCellbrowserConf, parseIntoColumns, readMatrixAnndata, splitOnce
from .cellbrowser import generateHtmls, writePyConf

from os.path import join, basename, dirname, isfile, isdir, relpath, abspath, getsize, getmtime, expanduser

def cbGet_parseArgs(showHelp=False):
    " setup logging, parse command line arguments and options. -h shows auto-generated help page "
    parser = optparse.OptionParser("""usage: %prog [options] <repo> <accession> - import single cell data from repositories

    The only valid repo right now is: ebi

    Examples:
    - %prog ebi -a E-GEOD-100058 -o myEbiDataset - import E-GEOD-10058 to myEbiDataset/
    - %prog ebi -i E-GEOD-100058-mirror/ -o myEbiDataset - convert E-GEOD-10058 FTP mirror to myEbiDataset/
    """)

    parser.add_option("-d", "--debug", dest="debug", action="store_true",
        help="show debug messages")
    parser.add_option("-a", "--accession", dest="acc", action="store",
        help="Accession in source database")
    parser.add_option("-i", "--inDir", dest="inDir", action="store",
        help="Input directory")
    parser.add_option("-o", "--outDir", dest="outDir", action="store",
        help="Output directory")

    (options, args) = parser.parse_args()

    if showHelp:
        parser.print_help()
        exit(1)

    setDebug(options.debug)
    return args, options

def mirrorFtp(hostName, ftpDir, localDir):
    " mirror an ftp directory to local disk "
    makeDir(localDir)
    from ftplib import FTP
    logging.info("FTP: connecting to %s, dir %s" % (hostName, ftpDir))
    ftp = FTP(hostName)
    ftp.login()
    ftp.cwd(ftpDir)

    ls = ftp.nlst()
    count = len(ls)
    curr = 0
    logging.info("Found {} files".format(count))
    for fn in ls:
        if fn=="complete":
            continue
        curr += 1
        logging.info('Downloading file {} ... {} of {} ...'.format(fn, curr, count))
        outPath = join(localDir, fn)
        ftp.retrbinary('RETR ' + fn, open(outPath, 'wb').write)

    ftp.quit()

    complFn = join(localDir, "complete")
    open(complFn, "w").close() # = create 0-size file
    logging.info("FTP download complete.")

#def downloadFromEbi(acc, outDir):
    #" download files to outDir/orig and then create cell browser files in outDir/ "
    #for fExt in [".clusters.tsv"]:
        #fname = acc+fExt
        #fileUrl = baseUrl+"/"+fname
        #localPath = join(outDir, fname)

def parseClusters(fname):
    """ find the optimal K number in the clusters file and the cluster assignment
    return a the value of K and a dictionary with sampleName -> clusterNumber
    """
    logging.info("Parsing %s" % fname)
    bestK = None
    for line in open(fname):
        row = line.rstrip("\n\r").split("\t")
        if row[0]=="sel.K":
            sampleNames = row[2:]
            continue
        selK = row[0]
        lastRow = row
        if selK=="TRUE":
            bestK = row[1]
            clusters = row[2:]
            break

    if bestK is None:
        logging.warn("No best K found, using last K")
        bestK = lastRow[1]
        clusters = lastK[2:]

    return bestK, dict(zip(sampleNames, clusters))

def parseCondSdrf(cellIds, fname):
    """ parse the condensed sdrf into cellId -> dict of key-val """
    cellIds = set(cellIds)
    cellMeta = defaultdict(dict)
    for line in open(fname):
        #E-MTAB-7303             ERR2847884      characteristic  organism        Homo sapiens    http://purl.obolibrary.org/obo/NCBITaxon_9606
        #E-MTAB-4850		SAMEA50486668	characteristic	age	42 year
        row = line.rstrip("\n\r").split("\t")
        dsId, _, cellId, _, fieldName, val = row[:6]
        if cellId not in cellIds:
            continue
        cellMeta[cellId][fieldName] = val
    return cellMeta

def parseSdrfFieldNames(sdrfFname):
    " return list of field names in sdrf file "
    ifh = open(sdrfFname)

    #Source Name	Comment[ENA_SAMPLE]	Comment[BioSD_SAMPLE]	Characteristics[organism]	Characteristics[disease]	Characteristics[disease staging]	Characteristics[age]	Unit[time unit]	Characteristics[sex]	Characteristics[organism part]	Characteristics[cell type]	Characteristics[cell line]	Characteristics[single cell well quality]	Comment[single cell quality]	Material Type	Protocol REF	Protocol REF	Extract Name	Comment[LIBRARY_LAYOUT]	Comment[LIBRARY_SELECTION]	Comment[LIBRARY_SOURCE]	Comment[LIBRARY_STRAND]	Comment[LIBRARY_STRATEGY]	Comment[NOMINAL_LENGTH]	Comment[NOMINAL_SDEV]	Comment[ORIENTATION]	Comment[single cell isolation]	Comment[library construction]	Comment[input molecule]	Comment[primer]	Comment[end bias]	Protocol REF	Performer	Assay Name	Comment[technical replicate group]	Technology Type	Comment[ENA_EXPERIMENT]	Scan Name	Comment[SUBMITTED_FILE_NAME]	Comment[ENA_RUN]	Comment[FASTQ_URI]	Comment[SPOT_LENGTH]	Comment[READ_INDEX_1_BASE_COORD]	Factor value[cell line]	Factor Value[stimulus]	Factor Value[single cell identifier]

    fieldNames = ifh.readline().strip("\n\r").split("\t")

    newNames = []
    for fieldName in fieldNames:
        if "[" in fieldName:
            fieldName = fieldName.split("[")[1].rstrip("]")
        newNames.append(fieldName)
    return newNames

#def parseSdrf(srdfFname):
    #""" parse EBI srdf file to list of fieldNames. Remove fields with only a single value.
    #return a dict with the fields and values where all lines contain the same value.
    #Also remove filename and URI fields.
    #"""
    #logging.debug("Parsing %s" % srdfFname)
    #columns = parseIntoColumns(srdfFname)
    ##cleanList = []
    #fieldList = []
    #genInfo = {}
    #for fieldName, values in columns:
        #if "[" in fieldName:
            #fieldName = fieldName.split("[")[1].rstrip("]")

        #valSet = set(values)

        #logging.debug(fieldName)
        #logging.debug(valSet)
        #if len(valSet)==1:
            #logging.debug("Field has only a single value, adding to info dict: %s=%s" % (fieldName, values[0]))
            #genInfo[fieldName] = values[0]
        #else:
            #cleanList.append( (fieldName, values) )
            #fieldList.append(fieldName)

    #return fieldList, genInfo

def parseIdf(fname):
    " parse idf and return as dict key = value "
    logging.debug("Parsing %s" % fname)
    data = OrderedDict()
    for line in open(fname, "Ut"):
        if line.startswith("\n"):
            continue
        line = line.rstrip("\n")
        key, val = splitOnce(line, "\t")
        data[key] = val
    return data

def toStr(s):
    " prettify raw IDF tab-sep list to a normal string "
    return s.strip("\t").replace("\t", ", ")

def translateIdf(idf):
    " convert dict read from IDF to a simple key-value format for the cell browser desc.conf "
    desc = OrderedDict()
    desc["title"] = toStr(idf["Investigation Title"])
    desc["abstract"] = toStr(idf["Experiment Description"])
    desc["design"] = toStr(idf["Experimental Design"])
    desc["expType"] = toStr(idf["Comment[EAExperimentType]"])
    desc["methods"] = toStr(idf["Protocol Description"])
    desc["curator"] = toStr(idf["Comment[EACurator]"])+", EBI Single Cell Expression Atlas"
    desc["submitter"] = toStr(idf["Person First Name"])+" "+toStr(idf["Person Last Name"])+" <"+toStr(idf["Person Email"])+">"
    desc["submission_date"] = toStr(idf["Public Release Date"])
    if "PubMed ID" in idf:
        desc["pmid"] = toStr(idf["PubMed ID"])
    if "Publication DOI" in idf:
        desc["doi"] = toStr(idf["Publication DOI"])

    if "Comment[ArrayExpressAccession]" in idf:
        desc["arrayexpress"] = toStr(idf["Comment[ArrayExpressAccession]"])

    if "Comment [SecondaryAccession]" in idf:
        desc["ena_project"] = toStr(idf["Comment [SecondaryAccession]"])

    return desc

def convertCoords(coordFname, newCoordFname):
    " simply reorder columns"
    ofh = open(newCoordFname, "w")
    for line in open(coordFname):
        row = line.rstrip("\n\r").split("\t")
        newRow = (row[2], row[0], row[1])
        ofh.write("\t".join(newRow))
        ofh.write("\n")
    ofh.close()
    logging.info("Wrote %s" % newCoordFname)

def convertMarkers(inFname, outFname):
    " reorder columns and reheader and return list of one top marker per cluster "
    #names	groups	scores	logfc	pvals	pvals_adj
    #ENSG00000112096	0	9.120421	1.4149776	3.5174541399729827e-15	5.52275474517158e-11
    ofh = open(outFname, "w")
    clusterToMarkers = defaultdict(list)
    for line in open(inFname):
        row = line.rstrip("\n\r").split("\t")
        #cluster, pval, auroc, gene = row
        geneId, cluster, score, logfc, pval, pvalAdj = row
        if row[0]=="names":
            newRow = ["Cluster", "Gene", "p-Value Adj.", "p-Value", "logFC", "score"]
        else:
            newRow = (row[1], row[0], row[5], row[4], row[3], row[2])
            clusterToMarkers[cluster].append( (float(pvalAdj), geneId) )

        ofh.write("\t".join(newRow))
        ofh.write("\n")

    ofh.close()
    logging.info("Wrote %s" % outFname)

    topMarkers = {}
    for cluster, pValGenes in iterItems(clusterToMarkers):
        pValGenes.sort()
        topGene = pValGenes[0][1]
        topMarkers[cluster] = topGene

    return topMarkers

def writeQuickGenes(topGenes, outFname):
    " given cluster -> gene, create a quickGenes file "
    ofh = open(outFname, "w")
    for cluster, gene in iterItems(topGenes):
        ofh.write("%s\tCluster %s\n"% (gene, cluster))
    ofh.close()
    logging.info("Wrote %s" % outFname)

def makeBasicCbConf(name, shortLabel, hasInferred, matrixFname="exprMatrix.tsv.gz"):
    " just fill a basic conf dict and return it "
    c = OrderedDict()
    c["name"] = name
    c["shortLabel"] = shortLabel
    c["exprMatrix"] = matrixFname
    c["meta"] = "meta.tsv"
    c["#priority"] = "10"
    c["tags"] = ["ebi"]
    c["enumFields"] = "cluster"

    if hasInferred:
        clusterField = "inferred cell type"
        c["acroFname"] = "acronyms.tsv"
    else:
        clusterField = "cluster"

    c["clusterField"] = clusterField
    c["labelField"] = clusterField

    c["#unit"] = "TPM"
    c["coords"] = [{"file":"tsne_perp25.coords.tsv", "shortLabel" : "t-SNE Perp=25"}]
    c["markers"] = [{"file":"markers.tsv", "shortLabel":"Cluster-specific genes"}]
    c["quickGenesFile"] = "quickGenes.tsv"
    return c

def writeMeta(clusters, cellIds, sdrfFields, boringFields, cellMeta, metaFname):
    " write a meta.tsv file, add in the clusters "
    #idIdx = None
    #fieldNames = []
    #for fieldIdx, (fieldName, vals) in enumerate(metaColumns):
        #if fieldName=="ENA_RUN":
            #idIdx = fieldIdx
            #break
        #fieldNames.append(fieldName)

    #if idIdx is None:
        #errAbort("Couldn't find the main ID field in the sdrf file. Possible field names: %s" % ",".join(fieldNames))

    # basic check of the ID field
    #idName = metaColumns[idIdx][0]
    #idList = metaColumns[idIdx][1]
    #if len(idList)!=len(set(idList)):
        #idCounts = Counter(idList)
        #errAbort("field %s was identified as the ID field, but it is not unique" % idCounts)

    # put the id field first
    #newOrder = []
    #newOrder.append(metaColumns[idIdx])
    #for i in range(0, idList):
        #if i==idIdx or i==0: # remove the first source name field
            #continue
        #newOrder.append(metaColumns[i])

    # add the cluster field
    #clustVals = []
    #for cellId in newOrder[0][1]:
        #clustId = clusters[cellId]
        #clustVals.append(clustId)

    #newOrder.insert(1, ("Cluster", clustVals))

    #headers = [f[0] for f in newOrder]

    #fieldValues = [x[1] for x in newOrder]
    #for i in range(0, len(idList)):
        #row = []
        #for field in fieldValues:
            #row.append(field[i])
        #ofh.write("\t".join(row))
        #ofh.write("\n")
    #ofh.close()
    #logging.info("Wrote %s" % newCoordFname)

    logging.debug("Fields in the SDRF: %s" % sdrfFields)
    logging.debug("Boring fields: %s" % boringFields.keys())

    allFields = set()
    for cellId in cellIds:
        meta = cellMeta[cellId]
        allFields.update(meta.keys())
    logging.debug("All fields from the condensed: %s" % allFields)

    fieldOrder = []
    doneFields = set()
    for fn in sdrfFields:
        if fn not in boringFields and fn in allFields and fn not in doneFields:
            fieldOrder.append(fn)
            doneFields.add(fn)

    logging.debug("final field order: %s" % fieldOrder)
    assert(len(set(fieldOrder))==len(fieldOrder)) # fields must not appear twice

    ofh = open(metaFname, "w")
    ofh.write("cell ID\tcluster\t")
    ofh.write("\t".join(fieldOrder))
    ofh.write("\n")

    for cellId, meta in iterItems(cellMeta):
        row = [cellId]
        clustId = clusters.get(cellId, "Unknown")
        row.append(clustId)
        for fieldName in fieldOrder:
            val = meta.get(fieldName, "")
            row.append(val)
        ofh.write("\t".join(row))
        ofh.write("\n")
    ofh.close()
    logging.info("Wrote %s" % ofh.name)

def findBoringFields(cellMetaDict):
    """ return a dict of all fieldNames that have only a single value. The keys of this dict are the field names, 
    the values of this dict are the single values."""
    fieldValues = defaultdict(set)
    for cellId, cellMeta in iterItems(cellMetaDict):
        for fieldName, val in iterItems(cellMeta):
            fieldValues[fieldName].add(val)

    boringFields = {}
    for fieldName, vals in iterItems(fieldValues):
        if fieldName.endswith("_FILE_NAME") or fieldName.endswith("_URI") or fieldName=="single cell identifier":
            logging.debug("Skipping field %s" % fieldName)
            boringFields[fieldName] = None
            continue
        if len(vals)==1:
            boringFields[fieldName] = list(vals)[0]
    return boringFields

def writeAcronyms(cellMeta, acroFname):
    " extract acronyms from inferred cell name field "
    ofh = open(acroFname, "w")
    for cellId, metaDict in iterItems(cellMeta):
        cellName = metaDict.get("inferred cell type")
        if cellName and "(" in cellName:
            longName, shortName = cellName.split("(")
            longName = longName.strip()
            shortName = shortName.strip(")")
            metaDict["inferred cell type"] = shortName
            ofh.write(shortName+"\t"+longName+"\n")
    ofh.close()
    logging.info("Wrote %s" % ofh.name)
    return cellMeta

def convertFromEbi(origDir, acc, outDir):
    " convert single cell expression files to cell browser format "
    # cluster is a special meta data attribute, parse it into a dict
    clusterFname = join(origDir, acc+".clusters.tsv")
    bestK, clusters = parseClusters(clusterFname)

    # best K determines marker file name
    markerFname = join(origDir, acc+".marker_genes_%s.tsv" % bestK)
    newMarkerFname = join(outDir, "markers.tsv")
    topGenes = convertMarkers(markerFname, newMarkerFname)

    quickGeneFname = join(outDir, "quickGenes.tsv")
    writeQuickGenes(topGenes, quickGeneFname)

    # make sure we use the same cellID order in meta and matrix
    sampleNameFname = join(origDir, acc+".aggregated_filtered_counts.mtx_cols")
    sampleNames = open(sampleNameFname).read().splitlines()

    # meta data
    sdrfFname = join(origDir, acc+".sdrf.txt")
    #fieldNames, sdrfInfo = parseSdrf(sdrfFname)
    fieldNames = parseSdrfFieldNames(sdrfFname) # we're doing this to get them in the proper order
    condSdrfFname = join(origDir, acc+".condensed-sdrf.tsv")
    cellMeta = parseCondSdrf(sampleNames, condSdrfFname)
    boringFields = findBoringFields(cellMeta)

    # default perp is 25
    coordFname = join(origDir, acc+".tsne_perp_25.tsv")
    newCoordFname = join(outDir, "tsne_perp25.coords.tsv")
    convertCoords(coordFname, newCoordFname)

    # dataset descriptor
    idfFname = join(origDir, acc+".idf.txt")
    idfInfo = parseIdf(idfFname)
    datasetDesc = translateIdf(idfInfo)
    datasetDesc["custom"] = boringFields
    descFname = join(outDir, "desc.conf")
    writePyConf(datasetDesc, descFname)

    # write acronyms and add the short names to the meta data
    hasInferred = ("inferred cell type" in fieldNames)
    if hasInferred:
        acroFname = join(outDir, "acronyms.tsv")
        cellMeta = writeAcronyms(cellMeta, acroFname)

    # write the meta 
    metaFname = join(outDir, "meta.tsv")
    writeMeta(clusters, sampleNames, fieldNames, boringFields, cellMeta, metaFname)

    # cellbrowser.conf file
    confFname = join(outDir, "cellbrowser.conf")
    conf = makeBasicCbConf(acc, datasetDesc["title"], hasInferred)
    writePyConf(conf, confFname)

    mtxName = join(origDir, acc+".aggregated_filtered_counts.mtx.gz")
    sampleNamesFname = join(origDir, acc+".aggregated_filtered_counts.mtx_cols")
    geneNamesFname = join(origDir, acc+".aggregated_filtered_counts.decorated.mtx_rows")
    outMatName = join(outDir, "exprMatrix.tsv.gz")
    #runGzip(inMatName, outMatName)
    if not isfile(outMatName):
        mtxToTsvGz(mtxName, geneNamesFname, sampleNamesFname, outMatName)

def cbGetCli():
    " run downloaders "
    args, options = cbGet_parseArgs()

    if len(args)<1:
        cbGet_parseArgs(showHelp=True)
        sys.exit(1)

    outDir = options.outDir
    if outDir is None:
        outDir = acc

    cmd = args[0]

    cmds = ["ebi"]

    if cmd=="ebi":
        acc = options.acc
        if options.inDir:
            origDir = options.inDir
        else:
            origDir = join(outDir, "orig")
            flagFname = join(origDir, "complete")
            logging.debug("Checking %s" % flagFname)
            if not isfile(flagFname):
                hostName = "ftp.ebi.ac.uk"
                ftpDir = "/pub/databases/microarray/data/atlas/sc_experiments/%s" % acc
                mirrorFtp(hostName, ftpDir, origDir)
        convertFromEbi(origDir, acc, outDir)
    else:
        errAbort("Repository %s is not valid. Valid repos are: %s" % (cmd, ", ".join(cmds)))



# annotate a list of gene IDs with links to various external databases

import logging, sys, optparse, re, unicodedata, string, csv
from collections import defaultdict, namedtuple
from os.path import join, basename, dirname, isfile

from .cellbrowser import openStaticFile, staticFileNextRow, openFile, splitOnce, iterItems, lineFileNextRow, setDebug
from .cellbrowser import readGeneSymbols

dataDir = "geneAnnot"

# no spaces/special chars in filenames - otherwise the URL will be rejected as invalid by urllib2
HPRD = join(dataDir, "HPRD_molecular_class_081914.txt")
HGNC = join(dataDir, "hgnc_complete_set_05Dec17.txt")
SFARI = join(dataDir, "SFARI-Gene_genes_export06-12-2017.csv")
OMIM = join(dataDir, "mim2gene.txt")
COSMIC = join(dataDir, "Census_allWed_Dec__6_18_35_54_2017.tsv")
HPO = join(dataDir, "hpo_frequent_7Dec17.txt")
BRAINSPANLMD = join(dataDir, "brainspan_genes.csv")
BRAINSPANMOUSEDEV = join(dataDir, "brainspanMouse_9Dec17.txt")
MGIORTHO = join(dataDir, "mgi_HGNC_homologene_8Dec17.txt")
EUREXPRESS = join(dataDir, "eurexpress_7Dec17.txt")
DDD = join(dataDir, "DDG2P_18_10_2018.csv.gz")


# ==== functions =====
    
def parseArgs():
    " setup logging, parse command line arguments and options. -h shows auto-generated help page "
    parser = optparse.OptionParser("""usage: %prog [options] inFname outFname - annotate a tab-sep gene list file with information from other databases
            
    A minimal input file has a header line with at one field called "gene" (=symbol or ENS/Entrez geneID) and
    one field called "cluster".
    
    In the cellbrowser, the cluster name should match the cluster name in the meta data file.""")

    parser.add_option("-d", "--debug", dest="debug", action="store_true", help="show debug messages")
    parser.add_option("", "--hprd", dest="hprd", action="store", help="location of HPRD file, default %default", default=HPRD)
    parser.add_option("", "--hgnc", dest="hgnc", action="store", help="location of HGNC file, default %default", default=HGNC)
    parser.add_option("", "--sfari", dest="sfari", action="store", help="location of SFARI file, default %default", default=SFARI)
    parser.add_option("", "--omim", dest="omim", action="store", help="location of OMIM file, default %default", default=OMIM)
    parser.add_option("", "--cosmic", dest="cosmic", action="store", help="location of COSMIC Census file, default %default", default=COSMIC)
    parser.add_option("", "--hpo", dest="hpo", action="store", help="location of HPO gene/disease/phenotype file, default %default", default=HPO)
    parser.add_option("", "--lmd", dest="lmd", action="store", help="location of BrainSpan LMD file, default %default", default=BRAINSPANLMD)
    parser.add_option("", "--mgiOrtho", dest="mgiOrtho", action="store", help="location of MGI Homologene file, default %default", default=MGIORTHO)
    parser.add_option("", "--eurexpress", dest="eurexpress", action="store", help="location of Eurexpress file, default %default", default=EUREXPRESS)
    parser.add_option("", "--brainspanMouseDev", dest="brainspanMouseDev", action="store", help="location of brainspan Mouse Development ISH file, default %default", default=BRAINSPANMOUSEDEV)
    #parser.add_option("-f", "--file", dest="file", action="store", help="run on file") 
    #parser.add_option("", "--test", dest="test", action="store_true", help="do something") 
    (options, args) = parser.parse_args()

    if args==[]:
        parser.print_help()
        exit(1)

    setDebug(options.debug)
    return args, options

# ----------- main --------------
def parseBrainspanLmd(inFname):
    " return entrez -> brainspanGeneId with all entrez IDs that are in the brainspan LMD set "
    with openStaticFile(inFname, 'r') as csvfile:
        ret = {}
        cr = csv.reader(csvfile)
        headers = None
        for row in cr:
            if headers == None:
                assert(row==['row_num','gene_id','ensembl_gene_id','gene_symbol','entrez_id'])
                headers = row
                continue
            ret[row[4]] = row[1]
    return ret


def parseSfari(inFname):
    " return symbol -> class "
    classDesc = {
            "" : "Autism, No category",
            "S" : "Autism, S - Syndromic",
            "1" : "Autism, 1 - High confidence",
            "2" : "Autism, 2 - Strong candidate",
            "3" : "Autism, 3 - Suggestive evidence",
            "4" : "Autism, 4 - Minimal evidence",
            "5" : "Autism, 5 - Hypothesized but untested",
            "6" : "Autism, 6 - Evidence does not support role",
    }
    # ['status', 'gene-symbol', 'gene-name', 'chromosome', 'genetic-category', 'gene-score', 'syndromic', 'number-of-reports']
    headers = None
    ret = {}
    with openStaticFile(inFname, 'r') as csvfile:
        spamreader = csv.reader(csvfile)
        for row in spamreader:
            if headers == None:
                headers = row
                assert(headers[5]=="gene-score")
                continue
            score = row[5]
            ret[row[1]] = classDesc[score]
    return ret

def parseHprd(inFname):
    " return entrez gene ID -> mol class from HPRD "
    ret = {}
    for row in staticFileNextRow(inFname):
        ret[row.Entrez_gene_ID] = row.Molecular_class
    return ret

def parseHgnc(inFname):
    " return dict with symbol -> entrez gene ID and hgncID to entrezId"
    ret = {}
    hgncIdToEntrez = {}
    for row in staticFileNextRow(inFname):
        entrezIds = row.entrez_id.strip('"').split("|")
        for entrezId in entrezIds:
            ret[row.symbol] = entrezId
            hgncIdToEntrez[row.hgnc_id] = entrezId # there is a 1:1 coorespondance between HGNC and entrez gene Id, except two miRNAs in 2017
    return ret, hgncIdToEntrez

def parseOmim(inFname):
    " return entrezId -> OMIM ID, link is https://omim.org/entry/601542 "
    # 100660  gene    218     ALDH3A1 ENSG00000108602
    ret = {}
    for line in openStaticFile(inFname):
        if line.startswith("#"):
            continue
        row = line.rstrip("\n").split("\t")
        if row[2]=="":
            continue
        ret[row[2]] = row[0]
    return ret

def parseCosmic(inFname):
    " return entrezId -> cancer types "
    ret = {}
    for row in staticFileNextRow(inFname):
        entrez = row.Entrez_GeneId
        cancerSomDesc = row.Tumour_Types_Somatic_
        cancerGermDesc = row.Tumour_Types_Germline_
        otherSyn = row.Other_Syndrome
        dis = []
        dis.extend(cancerSomDesc.strip('"').split(","))
        dis.extend(cancerGermDesc.strip('"').split(","))
        dis.extend(otherSyn.strip('"').split(","))
        disStr = ", ".join([s.strip() for s in dis if s!=""])
        disStr = disStr.replace(";", ",") # ; is reserved
        ret[entrez] = disStr
    return ret

def parseHpo(inFname):
    " return entrezId -> human phenotypes "
    geneToNames = defaultdict(set)
    # ORPHA:349	FUCA1	2517	HP:0000943	Dysostosis multiplex
    for line in openStaticFile(inFname):
        if line.startswith("#"):
            continue
        row = line.rstrip("\n").split("\t")
        entrez = row[2]
        hpoName = row[4]
        geneToNames[entrez].add(hpoName)

    ret = {}
    for entrez, names in iterItems(geneToNames):
        names = list(names)
        names.sort()
        ret[entrez] = ", ".join(names)
    return ret

def parseMgiOrtho(hgncIdToEntrez, inFname):
    " return dict with mouse entrezId -> human entrez "
    ret = {}
    # MGI Accession ID        Marker Symbol   Marker Name     Feature Type    EntrezGene ID   NCBI Gene chromosome    NCBI Gene start NCBI Gene end   NCBI Gene strand       Ensembl Gene ID Ensembl Gene chromosome Ensembl Gene start      Ensembl Gene end        Ensembl Gene strand     VEGA Gene ID    VEGA Gene chromosome  VEGA Gene start  VEGA Gene end   VEGA Gene strand        CCDS IDs        HGNC ID HomoloGene ID
    humanToMouse = defaultdict(list)
    for row in staticFileNextRow(inFname):
        hgncIds = row.HGNC_ID
        if hgncIds=="null" or hgncIds=="":
            continue
        for hgncId in hgncIds.split("|"):
            humanEntrez = hgncIdToEntrez[hgncId]
            # saw only 5 duplicated entrezIDs on the mouse side
            mouseEntrez = row.EntrezGene_ID
            ret[mouseEntrez] = humanEntrez
            humanToMouse[humanEntrez].append(mouseEntrez)
    return ret, humanToMouse

def parseEurexpress(mouseEntrezToHumanEntrez, inFname):
    " return dict with human entrez -> (eurexpressId, annotationStr) "
    # Template ID     Gene Symbol     Assay ID        EMAP Term       Entrez ID       Theiler Stage
    entrezToTerms = defaultdict(set)
    entrezToEuroexpress = dict()
    skippedMouseIds = set()
    for row in staticFileNextRow(inFname):
        mouseEntrez = row.Entrez_ID
        humanEntrez = mouseEntrezToHumanEntrez.get(mouseEntrez)
        if humanEntrez==None:
            skippedMouseIds.add(mouseEntrez)
            continue
        if row.EMAP_Term!="":
            entrezToTerms[humanEntrez].add(row.EMAP_Term)
        entrezToEuroexpress[humanEntrez] = row.Assay_ID

    logging.info("Eurexpress mouse entrez IDs: %d mappable, %d not-mappable to human " % (len(entrezToEuroexpress),len(skippedMouseIds)))
    logging.debug("Eurexpress mouse: mouse entrez IDs not mappable to human: %s" % ",".join(skippedMouseIds))
    ret = {}
    for entrezId, terms in iterItems(entrezToTerms):
        eurexpId = entrezToEuroexpress[entrezId]
        ret[entrezId] = (eurexpId, ", ".join(sorted(list(terms))))

    return ret

def parseDDD(fname):
    " parse DDD phenotype file "
    ret = {}
    #for row in staticFileNextRow(inFname):
        #print row
    return ret
    
def parseSimpleMap(inFname):
    " parse simple tab-sep key-value file and return as dict "
    ret = {}
    for line in openStaticFile(inFname):
        row = line.rstrip("\n").split("\t")
        ret[row[0]] = row[1]
    return ret

def tabGeneAnnotate(inFname, symToEntrez, symToSfari, entrezToClass, entrezToOmim, entrezToCosmic, entrezToHpo, entrezToLmd, entrezToEuroexpress, humanToMouseEntrezList, mouseEntrezToBrainspanMouseDev):
    " "
    headers = None
    geneToSym = -1
    for row in lineFileNextRow(inFname):
        if headers is None:
            headers = list(row._fields)
            headers.append("_hprdClass")
            headers.append("_expr")
            headers.append("_geneLists")
            yield headers
        sym = row[1]
        if "|" in sym: # marker gene lists can carry geneId|symbol, strip the symbol in this case and re-convert below
            sym = sym.split("|")[0]
        if "." in sym: # remove Ensembl version identifier
            sym = sym.split(".")[0]

        # convert gene IDs to symbols
        if geneToSym is -1:
            geneToSym = readGeneSymbols(None, [sym])
        if geneToSym is not None:
            geneId = geneToSym.get(sym)
            if geneId is None:
                logging.debug("Cannot find NCBI Gene ID for symbol: %s" % sym)
                geneId = sym
            sym = geneId

        hprdClass = ""
        entrezId = symToEntrez.get(sym)

        if entrezId == None:
            logging.debug("Cannot find entrezId for symbol %s" % sym)

        # now summarize the presence/absence of this gene in various specialized gene lists:
        # OMIM, COSMIC, SFARI
        hprdClass = entrezToClass.get(entrezId, "")
        geneLists = []
        if sym!="":
            # SFARI
            sfariInfo = symToSfari.get(sym)
            if sfariInfo is not None:
                sfariInfo = "SFARI||"+sfariInfo
                geneLists.append(sfariInfo)

        if entrezId is not None:
            # OMIM
            omimId = entrezToOmim.get(entrezId)
            if omimId is not None:
                omimInfo = "OMIM|"+omimId
                geneLists.append(omimInfo)

            # COSMIC
            cosmicDesc = entrezToCosmic.get(entrezId)
            if cosmicDesc is not None:
                cosmicDesc = "COSMIC||"+cosmicDesc
                geneLists.append(cosmicDesc)

            # HPO
            hpoDesc = entrezToHpo.get(entrezId)
            if hpoDesc is not None:
                hpoDesc = "HPO|"+entrezId+"|"+hpoDesc
                geneLists.append(hpoDesc)

        # links to gene expression databases
        exprParts = []
        if entrezId is not None:
            if entrezId in entrezToLmd:
                exprParts.append("BrainSpLMD|"+entrezId)

            if entrezId in entrezToEuroexpress:
                eurExpId, annotStr = entrezToEuroexpress[entrezId]
                annotStr = annotStr.replace(";", ",")
                exprParts.append("Eurexp|"+eurExpId+"|"+annotStr)

            mouseEntrezList = humanToMouseEntrezList[entrezId]
            for mouseEntrez in mouseEntrezList:
                if mouseEntrez in mouseEntrezToBrainspanMouseDev:
                    exprParts.append("BrainSpMouseDev|"+mouseEntrezToBrainspanMouseDev[mouseEntrez])

        row = list(row)
        row[1] = sym # in case the original ID was a geneID, not a symbol

        row.append(hprdClass)
        row.append(";".join(exprParts))
        row.append(";".join(geneLists))

        yield row

def cbMarkerAnnotateCli():
    args, options = parseArgs()

    entrezToBrainspanMouseDev = parseSimpleMap(options.brainspanMouseDev)
    symToEntrez, hgncIdToEntrez = parseHgnc(options.hgnc)
    mouseEntrezToHumanEntrez, humanToMouseEntrezList = parseMgiOrtho(hgncIdToEntrez, options.mgiOrtho)

    entrezToEuroexpress = parseEurexpress(mouseEntrezToHumanEntrez, options.eurexpress)
    entrezToLmd = parseBrainspanLmd(options.lmd)
    entrezToHpo = parseHpo(options.hpo)
    entrezToCosmic = parseCosmic(options.cosmic)
    entrezToOmim = parseOmim(options.omim)
    symToSfari = parseSfari(options.sfari)
    entrezToClass = parseHprd(options.hprd)
    #symToDdd = parseDdd(DDD)

    filename = args[0]
    outFname = args[1]

    rowCount = 0
    ofh = open(outFname, "w")
    for row in tabGeneAnnotate(filename, symToEntrez, symToSfari, entrezToClass, entrezToOmim, entrezToCosmic, entrezToHpo, entrezToLmd, entrezToEuroexpress, humanToMouseEntrezList, entrezToBrainspanMouseDev):
        ofh.write("\t".join(row))
        ofh.write("\n")
        rowCount +=1
    ofh.close()

    logging.info("Annotated %d marker gene rows, output written to %s" % (rowCount, outFname))


